import pandas as pd
import numpy as np
from datetime import datetime
from datetime import date
import json
from ..db.database import BasicDatabaseConnector,DerivedDatabaseConnector
from ..data_api.raw import get_raw_cm_index_price_df, get_cxindex_index_price_df, get_yahoo_index_price_df, get_rq_index_price_df,get_rq_index_weight,get_rq_stock_valuation,get_stock_fin_fac
from ..data_api.basic import get_index_info, get_trading_day_list, get_index_price, get_fund_ret,get_fund_info,get_fund_nav

class Process_Derived():
    
    w1 = 5
    m1 = 20
    m3 = 60
    m6 = 121
    y1 = 242
    y3 = 242*2
    y5 = 242*5
    y10 = 242*10



    def upload_derived(self, df, table_name):
        df.to_sql(table_name, DerivedDatabaseConnector().get_engine(), index=False, if_exists='append')
    
    def alpha(self, fund_nav,d):
            return fund_nav['adjusted_net_value'] / fund_nav['adjusted_net_value'].shift(d)  - fund_nav['close'] / fund_nav['close'].shift(d)


    def index_return(self):
        #history
        index_df = get_index_info()
        index_list = [i for o,i in zip(index_df['order_book_id'],index_df['index_id']) if '.' in o]
        for i in index_list:
            
            df =  get_index_price([i])
            if df.empty:
                continue
            else:
                df['w1_ret'] = df['close'] / df['close'].shift(self.w1) -1
                df['m1_ret'] = df['close'] / df['close'].shift(self.m1) -1
                df['m3_ret'] = df['close'] / df['close'].shift(self.m3) -1
                df['m6_ret'] = df['close'] / df['close'].shift(self.m6) -1
                df['y1_ret'] = df['close'] / df['close'].shift(self.y1) -1
                df['y3_ret'] = df['close'] / df['close'].shift(self.y3) -1
                df['y5_ret'] = df['close'] / df['close'].shift(self.y5) -1
                df['y10_ret'] = df['close'] / df['close'].shift(self.y10) -1
                df['cumulative_ret'] = df['close'] / df['close'].values[0] -1
                res = []
                df = df.set_index('datetime')
                for i,d in zip(df.close, df.index):
                    begin_of_year = df.index[df.index >= date(d.year, 1, 1)].min()
                    res.append(i/df.loc[begin_of_year,'close'] - 1)
                df['this_y_ret'] = res
                df = df.drop(['id','volume','low','close','high','open','total_turnover','ret'], axis = 1)
                df['datetime'] = df.index
                self.upload_derived(df, 'index_return')
    
    def index_volatility(self):
        #history
        index_df = get_index_info()
        index_list = index_df['index_id'].tolist()
        for i in index_list:
            df =  get_index_price([i])
            if df.empty:
                continue
            else:
                df = df.set_index('datetime')
                df = df[['close']]
                df = pd.DataFrame(df['close'] / df['close'][0])
                df['w1_vol'] = df['close'].rolling(self.w1).std()
                df['m1_vol'] = df['close'].rolling(self.m1).std()
                df['m3_vol'] = df['close'].rolling(self.m3).std()
                df['m6_vol'] = df['close'].rolling(self.m6).std()
                df['y1_vol'] = df['close'].rolling(self.y1).std()
                df['y3_vol'] = df['close'].rolling(self.y3).std()
                df['y5_vol'] = df['close'].rolling(self.y5).std()
                df['y10_vol'] = df['close'].rolling(self.y10).std()

                cumu_vol = []
                this_y_vol = []
                for d in df.index:
                    begin_of_year = df.index[df.index >= date(d.year, 1, 1)].min()
                    this_y_vol.append(df.loc[begin_of_year:d,]['close'].std())
                    cumu_vol.append(df.loc[:d,]['close'].std())
                df['this_y_vol'] = this_y_vol
                df['cumulative_vol'] = cumu_vol
                df = df.drop(['close'], axis = 1)
                df['index_id'] = i
                df['datetime'] = df.index
                self.upload_derived(df, 'index_volatility')
    
    def fund_alpha(self):

        index_df = get_index_info()
        fund_info = get_fund_info()
        fund_dic = {f: i for f,i in zip(fund_info['fund_id'],fund_info['track_index']) if i != 'none'}
        for f,i in fund_dic.items():

            fund_nav = get_fund_nav([f]).drop_duplicates(subset='datetime')
            index = get_index_price([i])

            fund_nav = fund_nav[['adjusted_net_value','datetime']].set_index('datetime')
            index = index[['close','datetime']].set_index('datetime')
            fund_nav = fund_nav.join(index)

            fund_nav['fund_ret'] = fund_nav['adjusted_net_value']/ fund_nav['adjusted_net_value'].shift(1) - 1
            fund_nav['index_ret'] = fund_nav['close'] /  fund_nav['close'].shift(1) - 1 
            track_err = []
            this_y_alpha = []
            cumu_alpha = []
            for d in fund_nav.index:
                f_ret = fund_nav.loc[:d,'fund_ret']
                i_ret = fund_nav.loc[:d,'index_ret']
                t = fund_nav.index.tolist().index(d) +1
                track_err_i = np.sqrt(pow((f_ret - i_ret),2).sum() *self.y1/ t)
                track_err.append(track_err_i)

                begin_of_year = fund_nav.index[fund_nav.index >= date(d.year, 1, 1)].min()
                this_y_alpha_i = fund_nav.loc[d,'adjusted_net_value'] / fund_nav.loc[begin_of_year,'adjusted_net_value'] - fund_nav.loc[d,'close'] / fund_nav.loc[begin_of_year,'close']
                this_y_alpha.append(this_y_alpha_i)

                cumu_alpha_i = fund_nav.loc[d,'adjusted_net_value'] / fund_nav['adjusted_net_value'].values[0] - fund_nav.loc[d,'close'] / fund_nav['close'].values[0]
                cumu_alpha.append(cumu_alpha_i)

            fund_nav['track_err'] = track_err
            fund_nav['this_y_alpha'] = this_y_alpha
            fund_nav['cumulative_alpha'] = cumu_alpha

            fund_nav['w1_alpha'] = self.alpha(fund_nav,self.w1)
            fund_nav['m1_alpha'] = self.alpha(fund_nav,self.m1)
            fund_nav['m3_alpha'] = self.alpha(fund_nav,self.m3)
            fund_nav['m6_alpha'] = self.alpha(fund_nav,self.m6)
            fund_nav['y1_alpha'] = self.alpha(fund_nav,self.y1)
            fund_nav['y3_alpha'] = self.alpha(fund_nav,self.y3)
            fund_nav['y5_alpha'] = self.alpha(fund_nav,self.y5)
            fund_nav['y10_alpha'] = self.alpha(fund_nav,self.y10)
            fund_nav = fund_nav.drop(['adjusted_net_value','close','fund_ret','index_ret'], axis = 1)
            fund_nav['fund_id'] = f
            fund_nav['datetime'] = fund_nav.index
            self.upload_derived(fund_nav, 'fund_alpha')
            

    def index_valuation(self):
        index_df = get_index_info()
        index_df = index_df[index_df['tag_method'] != 'none']
        index_df = index_df[index_df['order_book_id'] != 'not_available']
        index_df = index_df[index_df['order_book_id'].map(lambda x : '.' in x )]
        start_date = date(2010,1,1)
        index_result = []
        for i in index_df.to_dict('records'):
            index_res = {}
            index_id = i['index_id']
            index_weight = get_rq_index_weight([index_id],start_date)
            print(index_id)
            # save index components weight
            res = {}
            stock_list = []
            for s,w,d  in zip(index_weight['stock_list'] , index_weight['weight_list'],index_weight['datetime']):
                s = json.loads(s)
                w = json.loads(w)
                res[d] = [{'stock_id':si , 'weights': wi} for si,wi in zip(s,w)]
                stock_list.extend(s)
            stock_list = list(set(stock_list))
            # if all components are not stock, next one 
            stock_list = [ _ for _ in stock_list if _[0] in ['0','3','6']]
            if len(stock_list) < 1:
                continue
                
            stock_val  = get_rq_stock_valuation(stock_list,start_date)
            stock_val2 = get_stock_fin_fac(stock_list,start_date)
            
            for d, ws in res.items():
                index_res[d] = {}
                stock_val_tmp = stock_val[stock_val['datetime'] == d].copy()
                stock_val_tmp2 = stock_val2[stock_val2['datetime'] == d].copy()
                
                if stock_val_tmp.empty:
                    break
                for index_i in ['pb_ratio_lf','pe_ratio_ttm','peg_ratio_ttm','dividend_yield_ttm']:
                
                    index_d_w = stock_val_tmp[['stock_id',index_i]].set_index('stock_id').join(pd.DataFrame(res[d]).set_index('stock_id'))
                    index_d_w = index_d_w.fillna(0)
                    index_res[d][index_i] =  (index_d_w[index_i] * index_d_w['weights']).sum()
                
                index_d_w = stock_val2[['stock_id','return_on_equity_ttm']].set_index('stock_id').join(pd.DataFrame(res[d]).set_index('stock_id'))
                index_d_w = index_d_w.fillna(0)
                index_res[d]['return_on_equity_ttm'] = (index_d_w['return_on_equity_ttm'] * index_d_w['weights']).sum()
                
                
            index_result_i = pd.DataFrame(index_res).T
            index_result_i['index_id'] = index_id
            index_result_i['pe_pct'] = index_result_i['pe_ratio_ttm'].rank(pct=True)
            index_result_i['pb_pct'] = index_result_i['pb_ratio_lf'].rank(pct=True)
            index_result_i = index_result_i.dropna().copy()
            index_i = index_result_i['index_id'].values[0]
            val_method = index_df[index_df['index_id'] == index_i]['tag_method'].values[0]
            if val_method == 'PE百分位':
                index_result_i['val_score'] = index_result_i['pe_pct'].copy()
            else:
                index_result_i['val_score'] = index_result_i['pb_pct'].copy()
            index_result_i['datetime'] = index_result_i.index
            index_result_i.columns = ['dy_ttm','pb_mrq','pe_ttm','peg_ttm','roe_ttm','index_id','pe_pct','pb_pct','val_score','datetime']
            self.upload_derived(index_result_i, 'index_valuation')
            print(index_id, 'finish')


    




if __name__ == '__main__':
    
    pp = Process_Derived()
    pp.index_valuation()


from sqlalchemy import CHAR, Column, Integer, DATE
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.dialects.mysql import DOUBLE

Base = declarative_base()


class FundDailyCollection(Base):
    """基金信息日度收集"""

    __tablename__ = 'fund_daily_collection'
    id = Column(Integer, primary_key=True)

    datetime = Column(CHAR(20))  # 时间
    fund_id = Column(CHAR(20))  # 基金ID
    order_book_id = Column(CHAR(20))  # 基金代码

    wind_class_I = Column(CHAR(64))  # Wind基金类型
    wind_class_II = Column(CHAR(64))  # Wind基金二级类型
    institution_rating = Column(CHAR(20))  # 机构评级
    found_to_now = Column(DOUBLE(asdecimal=False))  # 成立年限
    average_size = Column(DOUBLE(asdecimal=False))  # 平均规模
    exchange_status = Column(CHAR(20))  # 交易状态
    theme = Column(CHAR(20))  # 基金主题
    track_index = Column(CHAR(20)) # 跟踪指数

    unit_net_value = Column(DOUBLE(asdecimal=False))  # 单位净值
    acc_net_value = Column(DOUBLE(asdecimal=False))  # 累积单位净值
    adjusted_net_value = Column(DOUBLE(asdecimal=False))  # 复权净值
    found_date = Column(CHAR(20))  # 成立日期
    annualized_returns = Column(DOUBLE(asdecimal=False))  # 成立以来年化收益率
    annualized_risk = Column(DOUBLE(asdecimal=False))  # 成立以来年化风险
    information_ratio = Column(DOUBLE(asdecimal=False))  # 成立以来信息比率
    last_month_return = Column(DOUBLE(asdecimal=False))  # 近一月收益率
    last_six_month_return = Column(DOUBLE(asdecimal=False))
    last_three_month_return = Column(DOUBLE(asdecimal=False))  # 近一季度收益率
    last_twelve_month_return = Column(DOUBLE(asdecimal=False))  # 近一年收益率
    last_week_return = Column(DOUBLE(asdecimal=False))  # 近一周收益率
    year_to_date_return = Column(DOUBLE(asdecimal=False))  # 今年以来收益率
    to_date_return = Column(DOUBLE(asdecimal=False))  # 成立至今收益率
    sharp_ratio = Column(DOUBLE(asdecimal=False))  # 成立至今夏普比率
    max_drop_down = Column(DOUBLE(asdecimal=False))  # 成立至今最大回撤

    fund_manager = Column(CHAR(255))  # 基金经理
    company_name = Column(CHAR(64))  # 基金公司
    symbol = Column(CHAR(64)) # 基金名称
    benchmark = Column(CHAR(255)) # 业绩基准

    zs = Column(DOUBLE(asdecimal=False))  # 招商评级
    sh3 = Column(DOUBLE(asdecimal=False))  # 上海证券评级三年期
    sh5 = Column(DOUBLE(asdecimal=False))  # 上海证券评级五年期
    jajx = Column(DOUBLE(asdecimal=False))  # 济安金信评级

    track_err = Column(DOUBLE(asdecimal=False))  # 跟踪误差
    this_y_alpha = Column(DOUBLE(asdecimal=False))  # 今年以来超额收益
    cumulative_alpha = Column(DOUBLE(asdecimal=False))  # 成立以来超额收益
    w1_alpha = Column(DOUBLE(asdecimal=False))  # 近一周收益率
    m1_alpha = Column(DOUBLE(asdecimal=False))  # 近一月超额收益
    m3_alpha = Column(DOUBLE(asdecimal=False))  # 近三月超额收益
    m6_alpha = Column(DOUBLE(asdecimal=False))  # 近半年超额收益
    y1_alpha = Column(DOUBLE(asdecimal=False))  # 近一年超额收益
    y3_alpha = Column(DOUBLE(asdecimal=False))  # 近三年超额收益
    y5_alpha = Column(DOUBLE(asdecimal=False))  # 近五年超额收益
    y10_alpha = Column(DOUBLE(asdecimal=False))  # 近十年超额收益

    latest_size = Column(DOUBLE(asdecimal=False)) # 最新规模

    beta = Column(DOUBLE(asdecimal=False))  # 风险指数
    alpha = Column(DOUBLE(asdecimal=False))  # 投资回报
    tag_track_err = Column(DOUBLE(asdecimal=False))  # tag跟踪误差
    fee_rate = Column(DOUBLE(asdecimal=False))  # 费率

    tag_name = Column(CHAR(64)) #基金类别
    score = Column(DOUBLE(asdecimal=False)) #评分

    @staticmethod
    def trans_columns():
        return {
            'fund_id': '基金ID',
            'order_book_id': '基金代码',
            'wind_class_I': '基金类型',
            'symbol': '基金名称',
            'institution_rating': '机构评级',
            'found_to_now': '成立年限',
            'average_size': '基金规模',
            'track_index': '跟踪指数',

            'exchange_status': '交易状态',
            'theme': '基金主题',
            'adjusted_net_value': '累计净值',
            'unit_net_value': '净值',
            'found_date': '成立日期',
            'annualized_returns': '年化收益',
            'annualized_risk': '成立以来年化风险',
            'information_ratio': '成立以来信息比率',
            'last_week_return': '近1周',
            'last_month_return': '近1月',
            'last_three_month_return': '近3月',
            'last_six_month_return': '近半年',
            'last_twelve_month_return': '近1年',
            'year_to_date_return': '年初至今',
            'to_date_return': '成立至今',
            'sharp_ratio': '夏普比率',
            'max_drop_down': '最大回撤',
            'fund_manager': '基金经理',
            'company_name': '基金公司',
            'benchmark': '业绩基准',
            'zs': '招商评级',
            'sh3': '上证三年评级',
            'sh5': '上证五年评级',
            'jajx': '济安评级',

            'track_err': '跟踪误差',
            'this_y_alpha': '今年以来超额收益',
            'cumulative_alpha': '成立以来超额收益',
            'w1_alpha': '近一周收益率',
            'm1_alpha': '近一月超额收益',
            'm3_alpha': '近三月超额收益',
            'm6_alpha': '近半年超额收益',
            'y1_alpha': '近一年超额收益',
            'y3_alpha': '近三年超额收益',
            'y5_alpha': '近五年超额收益',
            'y10_alpha': '近十年超额收益',
            'latest_size': '最新规模',

            'beta': 'beta',
            'alpha': 'alpha',
            'tag_track_err': 'tag跟踪误差',
            'fee_rate': '费率',
            'tag_name': '基金类别',
            'score': '基金评分',
        }


class IndexDailyCollection(Base):
    """基金信息日度收集"""

    __tablename__ = 'index_daily_collection'
    id = Column(Integer, primary_key=True)

    index_id = Column(CHAR(20))  # 指数ID
    datetime = Column(DATE)  # 日期
    pb_mrq = Column(DOUBLE(asdecimal=False))  # 市净率-MRQ
    pe_ttm = Column(DOUBLE(asdecimal=False))  # 市盈率-MMT
    peg_ttm = Column(DOUBLE(asdecimal=False))  # PEG-MMT
    roe_ttm = Column(DOUBLE(asdecimal=False))  # 净资产收益率-MMT
    dy_ttm = Column(DOUBLE(asdecimal=False))  # 股息率-MMT
    pe_pct = Column(DOUBLE(asdecimal=False))  # PE百分位
    pb_pct = Column(DOUBLE(asdecimal=False))  # PB百分位
    val_score = Column(DOUBLE(asdecimal=False))  # 估值评分

    vol_datetime = Column(DATE)  # 日期
    w1_vol = Column(DOUBLE(asdecimal=False))  # 近一周波动率
    m1_vol = Column(DOUBLE(asdecimal=False))  # 近一月波动率
    m3_vol = Column(DOUBLE(asdecimal=False))  # 近三月波动率
    m6_vol = Column(DOUBLE(asdecimal=False))  # 近半年波动率
    y1_vol = Column(DOUBLE(asdecimal=False))  # 近一年波动率
    y3_vol = Column(DOUBLE(asdecimal=False))  # 近三年波动率
    y5_vol = Column(DOUBLE(asdecimal=False))  # 近五年波动率
    y10_vol = Column(DOUBLE(asdecimal=False))  # 近十年波动率
    this_y_vol = Column(DOUBLE(asdecimal=False))  # 今年以来波动率
    cumulative_vol = Column(DOUBLE(asdecimal=False))  # 成立至今波动率

    ret_datetime = Column(DATE)  # 日期
    w1_ret = Column(DOUBLE(asdecimal=False))  # 近一周收益率
    m1_ret = Column(DOUBLE(asdecimal=False))  # 近一月收益率
    m3_ret = Column(DOUBLE(asdecimal=False))  # 近三月收益率
    m6_ret = Column(DOUBLE(asdecimal=False))  # 近半年收益率
    y1_ret = Column(DOUBLE(asdecimal=False))  # 近一年收益率
    y3_ret = Column(DOUBLE(asdecimal=False))  # 近三年收益率
    y5_ret = Column(DOUBLE(asdecimal=False))  # 近五年收益率
    y10_ret = Column(DOUBLE(asdecimal=False))  # 近十年收益率
    this_y_ret = Column(DOUBLE(asdecimal=False))  # 今年以来收益率
    cumulative_ret = Column(DOUBLE(asdecimal=False))  # 成立至今收益率

    price_datetime = Column(DATE)  # 日期
    volume = Column(DOUBLE(asdecimal=False))  # 交易量
    low = Column(DOUBLE(asdecimal=False))  # 最低价
    close = Column(DOUBLE(asdecimal=False))  # 收盘价
    high = Column(DOUBLE(asdecimal=False))  # 最高价
    open = Column(DOUBLE(asdecimal=False))  # 开盘价
    total_turnover = Column(DOUBLE(asdecimal=False))  # 成交额
    ret = Column(DOUBLE(asdecimal=False))  # 收益率

    order_book_id = Column(CHAR(20))  # 米筐ID
    industry_tag = Column(CHAR(64))  # 行业标签
    tag_method = Column(CHAR(64))  # 估值评分采用方法
    desc_name = Column(CHAR(64))  # 名称

    @staticmethod
    def trans_columns():
        return {
            'index_id': '指数ID',
            'datetime': '日期',
            'w1_vol': '近一周波动率',
            'm1_vol': '近一月波动率',
            'm3_vol': '近三月波动率',
            'm6_vol': '近半年波动率',
            'y1_vol': '近一年波动率',
            'y3_vol': '近三年波动率',
            'y5_vol': '近五年波动率',
            'y10_vol': '近十年波动率',
            'this_y_vol': '今年以来波动率',
            'cumulative_vol': '成立至今波动率',

            'ret_datetime': 'ret日期',
            'w1_ret': '近一周收益率',
            'm1_ret': '近一月收益率',
            'm3_ret': '近三月收益率',
            'm6_ret': '近半年收益率',
            'y1_ret': '近一年收益率',
            'y3_ret': '近三年收益率',
            'y5_ret': '近五年收益率',
            'y10_ret': '近十年收益率',
            'this_y_ret': '今年以来收益率',
            'cumulative_ret': '成立至今收益率',

            'pb_mrq': 'PB',
            'pe_ttm': 'PE',
            'peg_ttm': '预测PEG',
            'roe_ttm': 'ROE',
            'dy_ttm': '股息率',
            'pe_pct': 'PE百分位',
            'pb_pct': 'PB百分位',
            'val_score': '估值评分',
            'alpha_datetime': 'alpha日期',

            'price_datetime': '价格日期',
            'volume': '交易量',
            'low': '最低价',
            'close': '收盘价',
            'high': '最高价',
            'open': '开盘价',
            'total_turnover': '成交额',
            'ret': '收益率',

            'order_book_id': '指数代码',
            'industry_tag': '行业标签',
            'tag_method': '估值评分采用方法',
            'desc_name': '指数名称',
        }






from .processor import IfaProcessor
from .struct import WxMsgItem

class FundProcessor(IfaProcessor):

    def __init__(self):
        pass

    def process(self, msg: WxMsgItem):
        if msg.msgtype == 'text':


from datetime import datetime
import sys
import os
from collections import defaultdict
import datetime
import numpy as np
import time
import json
import hashlib
import requests
from urllib.parse import urlencode

class Utility(object):

    @staticmethod
    def get_n_float(num, n):
        """
        只保留小数点后n位，不进行四舍五入
        :param num:
        :param n:
        :return:
        """
        base = pow(10, n)
        return int(num * base) / base



from setuptools import setup, find_packages
from surfing import __version__


setup(
    name='surfing',
    version=__version__,
    description='backtest engine',
    author='traders.link',
    author_email='info@puyuan.tech',
    packages=find_packages()
)

#python3 setup.py bdist_egg --exclude-source-files --dist-dir=../


from ..util.constant import ReadableEnum

class DataType(ReadableEnum):
    #Industry Info data
    INDUSTRY_INFO = 200
    #Industry p/e
    INDUSTRY_PE = 201
    # stock latest financial indicator
    STOCK_FINANCIAL_INDICATOR = 202
    # stock daily value
    STOCK_DAILY_VALUE = 203
    # stock latest annual net profit
    STOCK_ANNUAL_NET_PROFIT = 204

    # industry pe ratio to update
    UPDATE_INDUSTRY_PE = 500
    # stock latest annual net profit to update
    UPDATE_STOCK_ANNUAL_NET_PROFIT = 501







#!/usr/bin/env python
# coding: utf-8
import numpy as np
import pandas as pd
import rqdatac as rq
import datetime
import json
from ..db.database import RawDatabaseConnector
from ..db.raw_models import *
from ..util.constant import STOCK_VALUATION_FACTORS, FUND_INDICATORS
from ..data_api.basic_new import BasicDataApi

class RqRawDataDownloader(object):
    def __init__(self, rq_license):
        # Initialize RiceQuant API
        rq.init('license', rq_license, ('rqdatad-pro.ricequant.com', 16011))

        # Get all fund order_book_ids 
        self._fund_order_book_id_list = list(set(rq.fund.all_instruments()['order_book_id'].tolist()))

        # Get stock order_book_ids
        self._stock_order_book_id_list = rq.all_instruments(type='CS', market='cn')['order_book_id'].tolist()

        # Get index list
        index_df = BasicDataApi().get_index_info()
        self._index_list = [order_book_id for order_book_id in index_df['order_book_id'].tolist() 
            if '.' in order_book_id and order_book_id != 'not_available']
        self._future_list = [order_book_id for order_book_id in index_df['order_book_id'].tolist() 
            if '.' not in order_book_id and order_book_id != 'not_available']
        self._order_book_id_2_index_id = {order_book_id : index_id for order_book_id, index_id in 
            zip(index_df['order_book_id'], index_df['index_id']) if order_book_id != 'not_available'}

    def _upload_raw(self, df, table_name, if_exists='append'):
        print(table_name)
        # print(df)
        df.to_sql(table_name, RawDatabaseConnector().get_engine(), index=False, if_exists=if_exists)

    def _get_date_list(self, start_date, end_date):
        start_datetime = datetime.datetime.strptime(start_date, '%Y%m%d')
        end_datetime = datetime.datetime.strptime(end_date, '%Y%m%d')
        date_list = []
        while start_datetime <= end_datetime:
            date_list.append(start_datetime.strftime('%Y%m%d'))
            start_datetime += datetime.timedelta(days=1)
        return date_list

    def rq_fund_indicator(self, start_date, end_date):
        df = rq.fund.get_indicators(self._fund_order_book_id_list, start_date, end_date, 
            fields=FUND_INDICATORS)

        if df is not None:
            df.reset_index(inplace=True)
            self._upload_raw(df, RqFundIndicator.__table__.name)

    def rq_fund_nav(self, start_date, end_date):
        df_list = []

        count = 0
        for order_book_id in self._fund_order_book_id_list:
            count += 1
            if count % 100 == 0:
                print(f'[{RqFundNav.__table__.name}] downloaded {count} out of '
                    f'{len(self._fund_order_book_id_list)} funds')
            df = rq.fund.get_nav(order_book_id, start_date, end_date, expect_df=True)
            if df is None:
                continue
            df.reset_index(inplace=True)
            df_list.append(df)

        if len(df_list) > 0:
            df_all = pd.concat(df_list)
            self._upload_raw(df_all, RqFundNav.__table__.name)
            
    def rq_fund_size(self):
        instruments = []
        for instrument in rq.fund.instruments(self._fund_order_book_id_list):
            instruments.append({
                'order_book_id' : instrument.order_book_id, 
                'latest_size' : instrument.latest_size
            })
        df = pd.DataFrame(instruments)
        if df is not None:
            df['update_time'] = datetime.date.today()
            self._upload_raw(df, RqFundSize.__table__.name, if_exists='replace')

    def rq_index_price(self, start_date, end_date):
        df = rq.get_price(self._index_list, start_date, end_date, expect_df=True)
        if df is not None:
            df = df.reset_index().drop(['num_trades'], axis = 1)
            df['datetime'] = df['date']
            df = df.drop(['date'], axis = 1)
            self._upload_raw(df, RqIndexPrice.__table__.name)

        df = rq.get_price(self._future_list, start_date, end_date, expect_df=True)
        if df is not None:
            df = df.reset_index().drop(['limit_down', 'prev_settlement', 'limit_up', 'open_interest', 'settlement', 
                'dominant_id'], axis = 1)
            df['datetime'] = df['date']
            df = df.drop(['date'], axis = 1)
            self._upload_raw(df, RqIndexPrice.__table__.name)

    def rq_index_weight(self, start_date, end_date):
        date_list = self._get_date_list(start_date, end_date)
        for i in self._index_list:
            res = []
            for d in date_list:
                try:
                    df = rq.index_weights(i, date=d)
                    if df is not None:
                        dic = df.to_dict()
                        stock_list = [k for k, v in dic.items()]
                        weight_list = [round(v, 6) for k, v in dic.items()]
                        res.append({
                            'index_id': self._order_book_id_2_index_id[i],
                            'datetime': d,
                            'weight_list': json.dumps(weight_list),
                            'stock_list': json.dumps(stock_list),  
                        })
                except:
                    pass
            if len(res) > 0:
                df = pd.DataFrame(res)
                self._upload_raw(df, RqIndexWeight.__table__.name)

    def rq_stock_fin_fac(self, start_date, end_date):
        factor = 'return_on_equity_ttm'
        df = rq.get_factor(self._stock_order_book_id_list, factor=factor, start_date=start_date, end_date=end_date)
        to_convert_series = type(df) is pd.core.series.Series
        if to_convert_series:
            df = df.to_frame()
        df = df.stack().reset_index()
        df.columns = ['stock_id', 'datetime', factor] if to_convert_series else ['datetime', 'stock_id', factor]
        self._upload_raw(df, RqStockFinFac.__table__.name)

    def rq_stock_post_price(self, start_date, end_date):
        df = rq.get_price(self._stock_order_book_id_list, start_date=start_date, end_date=end_date, 
            adjust_type='post', expect_df=True)
        if df is not None:
            df = df.reset_index().rename(columns={'date': 'datetime'})
            self._upload_raw(df, RqStockPostPrice.__table__.name)

    def rq_stock_price(self, start_date, end_date):
        df = rq.get_price(self._stock_order_book_id_list, start_date=start_date, end_date=end_date, 
            adjust_type='none', expect_df=True)
        if df is not None:
            df = df.reset_index().rename(columns={'date': 'datetime'})
            self._upload_raw(df, RqStockPrice.__table__.name)

    def rq_stock_valuation(self, start_date, end_date):
        df = rq.get_factor(self._stock_order_book_id_list, STOCK_VALUATION_FACTORS, start_date=start_date, 
            end_date=end_date, expect_df=True)
        if df is not None:
            df = df.reset_index().rename(columns={'order_book_id': 'stock_id', 'date':'datetime'}).replace(
                [np.inf, -np.inf], np.nan)
            self._upload_raw(df, RqStockValuation.__table__.name)

    def trading_day_list(self, start_date, end_date):
        df = pd.DataFrame(rq.get_trading_dates(start_date=start_date, end_date=end_date, market='cn'), 
            columns=['datetime'])
        if df is not None:
            self._upload_raw(df, TradingDayList.__table__.name)

    def stock_info(self):
        order_book_ids = []
        for stock_order_book_id in self._stock_order_book_id_list:
            order_book_ids.append({
                'rq_id' : stock_order_book_id, 
                'stock_id' : stock_order_book_id
            })
        df = pd.DataFrame(order_book_ids)
        if df is not None:
            self._upload_raw(df, StockInfo.__table__.name, if_exists='replace')

    def fund_rating(self, start_date, end_date):
        try:
            df = rq.fund.get_ratings(self._fund_order_book_id_list)
            if df is not None:
                df.reset_index(inplace=True)
                df = df[(df['datetime'] >= start_date) & (df['datetime'] <= end_date)]
                self._upload_raw(df, FundRating.__table__.name)
        except Exception as e:
            print(e)

    def download_all(self, start_date, end_date):
        self.rq_fund_indicator(start_date, end_date)
        self.rq_fund_nav(start_date, end_date)
        self.rq_index_price(start_date, end_date)
        self.rq_index_weight(start_date, end_date)
        self.rq_stock_fin_fac(start_date, end_date)
        self.rq_stock_post_price(start_date, end_date)
        self.rq_stock_price(start_date, end_date)
        self.rq_stock_valuation(start_date, end_date)
        self.trading_day_list(start_date, end_date)
        self.fund_rating(start_date, end_date)
        self.stock_info()
        self.rq_fund_size()



from sqlalchemy import CHAR, Column, Integer, TEXT
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.dialects.mysql import DOUBLE, DATE, TINYINT

Base = declarative_base()



class WindFundInfo(Base):
    '''万徳基金信息表'''

    __tablename__ = 'wind_fund_info'
    id = Column(Integer, primary_key=True)

    wind_id = Column(CHAR(20)) # 基金代码
    desc_name = Column(CHAR(64)) # 基金名称
    full_name = Column(CHAR(255)) #基金全称
    start_date = Column(DATE) # 成立日期
    end_date = Column(DATE) # 结束日期
    benchmark = Column(CHAR(255)) # 业绩比较基准
    wind_class_1 = Column(CHAR(64)) # wind投资分类一级
    wind_class_2 = Column(CHAR(64)) # wind投资分类二级
    currency = Column(CHAR(20)) # 交易币种
    base_fund_id = Column(CHAR(20)) # 分级基金母基金代码
    is_structured = Column(TINYINT(1)) # 是否分级基金
    is_open = Column(TINYINT(1)) # 是否定期开放基金
    manager_id = Column(CHAR(255)) # 基金经理(历任)
    company_id = Column(CHAR(64)) # 基金公司


class RqStockPrice(Base):
    '''米筐股票不复权数据表'''

    __tablename__ = 'rq_stock_price'
    id = Column(Integer, primary_key=True)

    open = Column(DOUBLE(asdecimal=False)) # 开盘价
    close = Column(DOUBLE(asdecimal=False)) # 收盘价
    high = Column(DOUBLE(asdecimal=False)) # 最高价
    low = Column(DOUBLE(asdecimal=False)) # 最低价
    limit_up = Column(DOUBLE(asdecimal=False)) # 涨停价
    limit_down = Column(DOUBLE(asdecimal=False)) # 跌停价
    total_turnover = Column(DOUBLE(asdecimal=False)) # 交易额
    volume = Column(DOUBLE(asdecimal=False)) # 交易量  不复权和后复权的交易量不同
    num_trades = Column(DOUBLE(asdecimal=False)) # 交易笔数
    datetime = Column(DATE) # 日期
    order_book_id = Column(CHAR(20)) # 股票ID
    
    
class RqStockPostPrice(Base):
    '''米筐股票后复权数据表'''

    __tablename__ = 'rq_stock_post_price'
    id = Column(Integer, primary_key=True)

    open = Column(DOUBLE(asdecimal=False)) # 开盘价
    close = Column(DOUBLE(asdecimal=False)) # 收盘价
    high = Column(DOUBLE(asdecimal=False)) # 最高价
    low = Column(DOUBLE(asdecimal=False)) # 最低价
    limit_up = Column(DOUBLE(asdecimal=False)) # 涨停价
    limit_down = Column(DOUBLE(asdecimal=False)) # 跌停价
    total_turnover = Column(DOUBLE(asdecimal=False)) # 交易额
    volume = Column(DOUBLE(asdecimal=False)) # 交易量  不复权和后复权的交易量不同
    num_trades = Column(DOUBLE(asdecimal=False)) # 交易笔数
    datetime = Column(DATE) # 日期
    order_book_id = Column(CHAR(20)) # 股票ID


class RqFundNav(Base):
    '''米筐基金净值表'''

    __tablename__ = 'rq_fund_nav'
    order_book_id = Column(CHAR(10), primary_key=True) # 合约代码
    datetime = Column(DATE, primary_key=True) # 日期

    unit_net_value = Column(DOUBLE(asdecimal=False)) # 单位净值
    acc_net_value = Column(DOUBLE(asdecimal=False)) # 累积单位净值
    adjusted_net_value = Column(DOUBLE(asdecimal=False)) # 复权净值
    change_rate = Column(DOUBLE(asdecimal=False)) # 涨跌幅
    daily_profit = Column(DOUBLE(asdecimal=False)) # 每万元收益（日结型货币基金专用）
    weekly_yield = Column(DOUBLE(asdecimal=False)) # 7日年化收益率（日结型货币基金专用）
    redeem_status = Column(CHAR(10)) # 赎回状态，开放 - Open, 暂停 - Suspended, 限制大额申赎 - Limited, 封闭期 - Close
    subscribe_status = Column(CHAR(10)) # 订阅状态，开放 - Open, 暂停 - Suspended, 限制大额申赎 - Limited, 封闭期 - Close
    
    
class RqStockValuation(Base):
    '''
    股票每日估值
    LF, Last File 时效性最好
    LYR, Last Year Ratio 上市公司年报有审计要求，数据可靠性最高
    TTM, Trailing Twelve Months 时效性较好，滚动4个报告期计算，可避免某一期财报数据的偶然性
    '''

    __tablename__ = 'rq_stock_valuation'
    id = Column(Integer, primary_key=True)

    stock_id = Column(CHAR(20)) # 股票id
    datetime = Column(DATE) # 日期
    pe_ratio_lyr = Column(DOUBLE(asdecimal=False)) # 市盈率lyr
    pe_ratio_ttm = Column(DOUBLE(asdecimal=False)) # 市盈率ttm
    ep_ratio_lyr = Column(DOUBLE(asdecimal=False)) # 盈市率lyr
    ep_ratio_ttm = Column(DOUBLE(asdecimal=False)) # 盈市率ttm
    pcf_ratio_total_lyr = Column(DOUBLE(asdecimal=False)) # 市现率_总现金流lyr
    pcf_ratio_total_ttm = Column(DOUBLE(asdecimal=False)) # 市现率_总现金流ttm
    pcf_ratio_lyr = Column(DOUBLE(asdecimal=False)) # 市现率_经营lyr
    pcf_ratio_ttm = Column(DOUBLE(asdecimal=False)) # 市现率_经营 ttm
    cfp_ratio_lyr = Column(DOUBLE(asdecimal=False)) # 现金收益率lyr
    cfp_ratio_ttm = Column(DOUBLE(asdecimal=False)) # 现金收益率ttm
    pb_ratio_lyr = Column(DOUBLE(asdecimal=False)) # 市净率lyr
    pb_ratio_ttm = Column(DOUBLE(asdecimal=False)) # 市净率ttm
    pb_ratio_lf = Column(DOUBLE(asdecimal=False)) # 市净率lf
    book_to_market_ratio_lyr = Column(DOUBLE(asdecimal=False)) # 账面市值比lyr
    book_to_market_ratio_ttm = Column(DOUBLE(asdecimal=False)) # 账面市值比ttm
    book_to_market_ratio_lf = Column(DOUBLE(asdecimal=False)) # 账面市值比lf
    dividend_yield_ttm = Column(DOUBLE(asdecimal=False)) # 股息率ttm
    peg_ratio_lyr = Column(DOUBLE(asdecimal=False)) # PEG值lyr
    peg_ratio_ttm = Column(DOUBLE(asdecimal=False)) # PEG值ttm
    ps_ratio_lyr = Column(DOUBLE(asdecimal=False)) # 市销率lyr
    ps_ratio_ttm = Column(DOUBLE(asdecimal=False)) # 市销率ttm
    sp_ratio_lyr = Column(DOUBLE(asdecimal=False)) # 销售收益率lyr
    sp_ratio_ttm = Column(DOUBLE(asdecimal=False)) # 销售收益率ttm
    market_cap = Column(DOUBLE(asdecimal=False)) # 总市值1 
    market_cap_2 = Column(DOUBLE(asdecimal=False)) # 流通股总市值
    market_cap_3 = Column(DOUBLE(asdecimal=False)) # 总市值3
    a_share_market_val = Column(DOUBLE(asdecimal=False)) # A股市值
    a_share_market_val_in_circulation = Column(DOUBLE(asdecimal=False)) # 流通A股市值
    ev_lyr = Column(DOUBLE(asdecimal=False)) # 企业价值lyr
    ev_ttm = Column(DOUBLE(asdecimal=False)) # 企业价值ttm
    ev_lf = Column(DOUBLE(asdecimal=False)) # 企业价值lf 
    ev_no_cash_lyr = Column(DOUBLE(asdecimal=False)) #企业价值(不含货币资金)lyr
    ev_no_cash_ttm = Column(DOUBLE(asdecimal=False)) # 企业价值(不含货币资金)ttm
    ev_no_cash_lf = Column(DOUBLE(asdecimal=False)) # 企业价值(不含货币资金)lf
    ev_to_ebitda_lyr = Column(DOUBLE(asdecimal=False)) # 企业倍数lyr
    ev_to_ebitda_ttm = Column(DOUBLE(asdecimal=False)) # 企业倍数ttm
    ev_no_cash_to_ebit_lyr = Column(DOUBLE(asdecimal=False)) #企业倍数(不含货币资金)lyr
    ev_no_cash_to_ebit_ttm = Column(DOUBLE(asdecimal=False)) # 企业倍数(不含货币资金)ttm
    

class RqFundIndicator(Base):
    '''米筐基金指标'''

    __tablename__ = 'rq_fund_indicator'
    id = Column(Integer, primary_key=True)

    order_book_id = Column(CHAR(10)) # 原始基金ID
    datetime = Column(DATE) # 日期

    last_week_return= Column(DOUBLE(asdecimal=False)) # 近一周收益率
    last_month_return= Column(DOUBLE(asdecimal=False)) # 近一月收益率
    last_three_month_return= Column(DOUBLE(asdecimal=False)) # 近一季度收益率
    last_six_month_return= Column(DOUBLE(asdecimal=False)) # 近半年收益率
    last_twelve_month_return= Column(DOUBLE(asdecimal=False)) # 近一年收益率
    year_to_date_return= Column(DOUBLE(asdecimal=False)) # 今年以来收益率
    to_date_return= Column(DOUBLE(asdecimal=False)) # 成立至今收益率
    average_size= Column(DOUBLE(asdecimal=False)) # 平均规模
    annualized_returns= Column(DOUBLE(asdecimal=False)) # 成立以来年化收益率
    annualized_risk= Column(DOUBLE(asdecimal=False)) # 成立以来年化风险
    sharp_ratio= Column(DOUBLE(asdecimal=False)) # 成立以来夏普比率
    max_drop_down= Column(DOUBLE(asdecimal=False)) # 成立以来最大回撤
    information_ratio= Column(DOUBLE(asdecimal=False)) # 成立以来信息比率


class RqIndexPrice(Base):
    '''米筐指数数据'''

    __tablename__ = 'rq_index_price'
    id = Column(Integer, primary_key=True)

    datetime = Column(DATE) # 日期
    order_book_id = Column(CHAR(20)) # 米筐代码
    high = Column(DOUBLE(asdecimal=False)) #最高价
    open = Column(DOUBLE(asdecimal=False)) #开盘价
    total_turnover = Column(DOUBLE(asdecimal=False)) # 交易额
    low = Column(DOUBLE(asdecimal=False)) # 最低价
    close = Column(DOUBLE(asdecimal=False)) # 收盘价
    volume = Column(DOUBLE(asdecimal=False)) # 交易量

class FundFee(Base):
    '''基金费率表'''

    __tablename__ = 'fund_fee'
    id = Column(Integer, primary_key=True)

    desc_name = Column(CHAR(32)) # 名称
    manage_fee = Column(DOUBLE(asdecimal=False)) # 管理费
    trustee_fee = Column(DOUBLE(asdecimal=False)) # 托管费
    purchase_fee = Column(DOUBLE(asdecimal=False)) # 申购费
    redeem_fee = Column(DOUBLE(asdecimal=False)) # 赎回费
    note = Column(CHAR(64)) # 附加信息
    fund_id = Column(CHAR(20)) # 基金id


class CxindexIndexPrice(Base):
    '''中证指数数据'''

    __tablename__ = 'cxindex_index_price'
    id = Column(Integer, primary_key=True)

    index_id = Column(CHAR(20)) # 指数名称
    datetime = Column(DATE) # 日期
    open = Column(DOUBLE(asdecimal=False)) # 开盘价
    close = Column(DOUBLE(asdecimal=False)) # 收盘价
    high = Column(DOUBLE(asdecimal=False)) # 最高价
    low = Column(DOUBLE(asdecimal=False)) # 最低价
    volume = Column(DOUBLE(asdecimal=False)) # 交易量
    total_turnover = Column(DOUBLE(asdecimal=False)) # 交易额
    ret = Column(DOUBLE(asdecimal=False)) # 收益率
    

class YahooIndexPrice(Base):
    '''雅虎指数数据'''

    __tablename__ = 'yahoo_index_price'
    id = Column(Integer, primary_key=True)

    index_id = Column(CHAR(20)) # 指数名称
    datetime = Column(DATE) # 日期
    open = Column(DOUBLE(asdecimal=False)) # 开盘价
    close = Column(DOUBLE(asdecimal=False)) # 收盘价
    high = Column(DOUBLE(asdecimal=False)) # 最高价
    low = Column(DOUBLE(asdecimal=False)) # 最低价
    volume = Column(DOUBLE(asdecimal=False)) # 交易量
    total_turnover = Column(DOUBLE(asdecimal=False)) # 交易额
    ret = Column(DOUBLE(asdecimal=False)) # 收益率


class CmIndexPrice(Base):
    '''汇率数据'''

    __tablename__ = 'cm_index_price'
    datetime = Column(DATE, primary_key=True) # 日期

    usd_central_parity_rate = Column(DOUBLE(asdecimal=False)) # 美元人民币汇率中间价
    eur_central_parity_rate = Column(DOUBLE(asdecimal=False)) # 欧元人民币汇率中间价
    jpy_central_parity_rate = Column(DOUBLE(asdecimal=False)) # 日元人民币汇率中间价
    usd_cfets = Column(DOUBLE(asdecimal=False)) # 美元人民币市询价
    eur_cfets = Column(DOUBLE(asdecimal=False)) # 欧元人民币市询价
    jpy_cfets = Column(DOUBLE(asdecimal=False)) # 日元人民币市询价


class FundRating(Base):
    '''基金评级'''

    __tablename__ = 'fund_rating'
    id = Column(Integer, primary_key=True)

    order_book_id = Column(CHAR(10)) # 米筐基金id
    datetime = Column(DATE) # 日期
    zs = Column(DOUBLE(asdecimal=False)) # 招商评级
    sh3 = Column(DOUBLE(asdecimal=False)) # 上海证券评级三年期
    sh5 = Column(DOUBLE(asdecimal=False)) # 上海证券评级五年期
    jajx = Column(DOUBLE(asdecimal=False)) # 济安金信评级


class RqIndexWeight(Base):
    '''指数成分权重'''

    __tablename__ = 'rq_index_weight'
    id = Column(Integer, primary_key=True)

    datetime = Column(DATE) # 日期
    index_id = Column(CHAR(20)) # 股票
    weight_list = Column(TEXT) # 权重列表 json 格式str 两者顺位对应
    stock_list = Column(TEXT) # 股票列表 json 格式str 两者顺位对应


class RqStockFinFac(Base):
    '''米筐股票财务因子'''

    __tablename__ = 'rq_stock_fin_fac'
    id = Column(Integer, primary_key=True)

    datetime = Column(DATE) # 日期
    stock_id =  Column(CHAR(20)) # 股票
    return_on_equity_ttm = Column(DOUBLE(asdecimal=False)) # 净资产收益率ttm
    

class RqFundSize(Base):
    '''米筐基金最新规模'''
    
    __tablename__ = 'rq_fund_size'
    order_book_id = Column(CHAR(10), primary_key=True) # 米筐基金id
    latest_size = Column(DOUBLE(asdecimal=False)) # 最新规模
    update_time = Column(DATE) # 更新日期


class StockInfo(Base):
    '''股票信息表'''

    __tablename__ = 'rq_stock_info'
    stock_id = Column(CHAR(20), primary_key=True) # 股票ID
    rq_id = Column(CHAR(20)) # 米筐ID


class TradingDayList(Base):
    '''交易日列表'''

    __tablename__ = 'rq_trading_day_list'
    datetime = Column(DATE, primary_key=True)


class IndexValPct(Base):
    '''指数估值'''
    
    __tablename__ = 'index_valpct'
    
    id = Column(Integer, primary_key=True)
    datetime = Column(DATE) # 更新日期
    index_id = Column(CHAR(20)) # 指数
    pe_ttm  = Column(DOUBLE(asdecimal=False)) # pe
    pe_pct = Column(DOUBLE(asdecimal=False)) # pe 百分位
    pb =  Column(DOUBLE(asdecimal=False)) # pb
    pb_pct =  Column(DOUBLE(asdecimal=False)) # pb pct
    



from .util import Singleton
import platform
import json

CURRENT_PLATFORM = platform.system()
DEFAULT_CONFIG_PATH = '/shared/etc/ifa/config.json'

class RedisSetting(object):

    def __init__(self, data_json):
        assert 'host' in data_json, 'could not find host in config file'
        assert 'port' in data_json, 'could not find port in config file'
        assert 'password' in data_json, 'could not find password in config file'
        assert 'key' in data_json, 'could not find key in config file'

        self.host = data_json['host']
        self.port = data_json['port']
        self.db = data_json.get('db', 0)
        self.password = data_json['password']
        self.key = data_json['key']

class MessageSenderSetting(object):

    def __init__(self, data_json):
        assert 'corp_id' in data_json, 'could not find corp_id in config file'
        assert 'app_id' in data_json, 'could not find app_id in config file'
        assert 'app_secret' in data_json, 'could not find add_secret in config file'

        self.corp_id = data_json['corp_id']
        self.app_id = data_json['app_id']
        self.app_secret = data_json['app_secret']

class Configurator(metaclass=Singleton):

    def __init__(self, config_path=DEFAULT_CONFIG_PATH):
        self.config_path = config_path
        if CURRENT_PLATFORM == 'Darwin':
            import os
            curr_file_dir = os.path.split(os.path.realpath(__file__))[0]
            self.config_path = os.path.join(curr_file_dir, '../../etc/config.json')
        self.config = json.load(open(self.config_path))

    def get_redis_setting(self):
        settings = RedisSetting(self.config['redis'])
        return settings

    def get_message_sender_setting(self):
        settings = MessageSenderSetting(self.config['wxwork_app'])
        return settings



import pandas as pd
import numpy as np
import datetime
import copy
from pprint import pprint 
import matplotlib as mpl
import pylab as pl
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import matplotlib.colors as colors
mpl.rcParams['font.family'] = ['Heiti TC']

from ...data.manager.manager_fund import FundDataManager
from ...data.struct import AssetWeight, AssetPosition, AssetPrice, AssetValue
class BacktestInfo:
    
    '''
    save backtest history
    '''
    TRADING_DAYS_PER_YEAR = 242

    def __init__(self):
        pass

    def init(self):
        pass    

    def setup(self, saa: AssetWeight):
        self.saa_weight = saa.__dict__
        self.cash_history = {}
        self.position_history = {}
        self.market_price_history = {}
        self.pending_trade_history = {}
        self.weights_history = {}
        self.tactic_history = {}
        self.trade_history = {}
        self.rebalance_date = []
    
    def update(self, dt: datetime, cur_position:AssetPosition,cur_price:AssetPrice,pend_trades: list ):   
        self.cash_history[dt] = cur_position.cash
        self.position_history[dt] = cur_position.__dict__
        self.market_price_history[dt] = AssetValue(prices=cur_price, positions=cur_position).sum()
        self.pending_trade_history[dt] = pend_trades
        self.weights_history[dt] = AssetValue(prices=cur_price, positions=cur_position).get_weight()
        #self.tactic_history[dt] = 
        #self.trade_history[dt] = 
        #self.rebalance_date.append()      

    def result(self):
        self.mv_df = pd.DataFrame([ {'date':k, 'mv':v} for k,v in self.market_price_history.items()]).set_index('date')
        year = self.mv_df.shape[0]/self.TRADING_DAYS_PER_YEAR
        total_return = self.mv_df['mv'][-1] / self.mv_df['mv'][0]
        try:
            five_year_return = (self.mv_df['mv'][-1] / self.mv_df['mv'][-(5 * self.TRADING_DAYS_PER_YEAR)] - 1)
        except:
            five_year_return = five_year_return = (self.mv_df['mv'][-1] / self.mv_df['mv'][0] - 1)
        annualized_return = np.exp(np.log(total_return)/year) - 1
        annualized_volatiltity = (self.mv_df['mv'].shift(1) / self.mv_df['mv']).std() * np.sqrt((self.mv_df.shape[0] - 1) / year)
        sharpe = annualized_return / annualized_volatiltity
        mdd = 1 - (self.mv_df.loc[:, 'mv'] / self.mv_df.loc[:, 'mv'].rolling(10000, min_periods=1).max()).min()
        mdd_part1 = (self.mv_df.loc[:, 'mv'] / self.mv_df.loc[:, 'mv'].rolling(10000, min_periods=1).max())
        self.mdd_date1 = self.mv_df.loc[:mdd_part1.idxmin(),'mv'].idxmax()
        self.mdd_date2 = mdd_part1.idxmin()
        w = copy.deepcopy(self.saa_weight)
        w['mdd'] = mdd
        w['annual_ret'] = annualized_return
        w['sharpe'] = sharpe
        w['5_year_ret'] = five_year_return
        w['annual_vol'] = annualized_volatiltity
        w['mdd_d1'] = self.mdd_date1
        w['mdd_d2'] = self.mdd_date2
        pprint(w)
        return w
    
    def backtest_plot(self, index_df):
        self.plot_asset_weights()
        self.plot_market_value(index_df)
        self.plot_mdd_period_assets(index_df)

    def plot_asset_weights(self):
        res = []
        for k,v in self.weights_history.items():
            v = v.__dict__
            v['date'] = k
            res.append(v)
        weights_df = pd.DataFrame(res).set_index('date')
        weights_df = weights_df.drop(['cash'], axis = 1).dropna()[1:]
        res = []
        for dic, s in zip(weights_df.to_dict('records'), weights_df.sum(axis = 1).values):
            res.append({ k: v/ s for k, v in dic.items()})
        df = pd.DataFrame(res)
        df.index = weights_df.index
        df.plot.area(figsize=(18,9),legend=False,fontsize = 13)
        l = pl.legend(loc='lower left',fontsize = 9)
        s = pl.title('weights history', fontsize=20)

    def plot_market_value(self, index_df):    
        index_df_raw = index_df.loc[self.mv_df.index[0]:self.mv_df.index[-1],]
        index_df = index_df_raw.copy().fillna(method='bfill')
        index_df = index_df/index_df.iloc[0]
        w_l = []
        for idx, r in index_df_raw.iterrows():
            nan_asset = [k for k,v in r.to_dict().items() if np.isnan(v)]
            wgts = self.saa_weight.copy()
            wgts['cash'] = 0
            for k in nan_asset:
                wgts[k] = 0
            s = sum([v  for k,v in wgts.items()])
            wgts = {k :v /s for k, v in wgts.items()}
            wgts['datetime'] = idx
            w_l.append(wgts)
        wgts_df = pd.DataFrame(w_l).set_index('datetime').drop(['cash'], axis = 1)
        self.mv_df['benchmark'] = (wgts_df * index_df).sum(axis = 1)
        self.mv_df =self.mv_df / self.mv_df.iloc[0]
        self.mv_df.plot.line(figsize=(20,12),legend=False,fontsize = 13)
        l = pl.legend(loc='lower left',fontsize = 10)
        s = pl.title('market value', fontsize=20)

    def plot_mdd_period_assets(self, index_df): 
        l = list(self.saa_weight.keys())
        l.remove('cash')
        df = index_df[l].loc[self.mdd_date1:self.mdd_date2].fillna(method='bfill')
        (df /df.iloc[0]).plot.line(figsize=(18,9),legend=False,fontsize = 13)
        l = pl.legend(loc='lower left',fontsize = 13)
        s = pl.title('during mdd period  market value of all assets', fontsize=20)    